/*  *************************************************************************

	Shadow Door - a Neverwinter Nights NPC control interface
	Copyright 2003 by Robert Zubek and Phillip Saltzman, Northwestern University
    email: rob@cs.northwestern.edu, p-saltzman@northwestern.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

************************************************************************** */


// the pointer table class
//
// the underlying structure is that of a dictionary. each node
// contains an entry which points to the value, and an optional
// dispatch table which points to other entries.

#if _MSC_VER > 1000
#pragma once
#endif 

#ifndef SD_POINTERTABLE_H
#define SD_POINTERTABLE_H

#include "definitions.h"

class Node;
typedef Node* NodePtr;

class Node {
protected:
	NodePtr *nodetable;
	ShadowDoorFnPtr entry;
	const int tablesize;

public:

	// constructor 
	Node () : tablesize(sizeof(char) * 256) 
	{
		nodetable = NULL;
		entry = NULL; 
	};

	// destructor: delete all next nodes, and the table.
	// DOES NOT delete the entry.
	virtual ~Node () 
	{
		if (nodetable) {
			for (char i = 0; i < tablesize; i++) delete (nodetable[i]);
			delete[] nodetable;
		}
	}

	// AddNode: adds the next node to the pointer table at 
	// a given index. allocates a new pointer table if necessary.
	inline void AddNode (char index, Node *next) 
	{
		if (!nodetable) {
			nodetable = new NodePtr[tablesize];
			ZeroMemory (nodetable, sizeof(NodePtr) * tablesize);
		}
		nodetable[index] = next;
	}

	// GetNode retrieves the next node at a given index.
	// returns NULL if the table entry does not exist.
	inline Node* GetNode (char index) 
	{
		if (!nodetable) return NULL;
		else return nodetable[index];
	}

	// AddEntry: adds a new pointer to the current node
	inline void AddPointer (ShadowDoorFnPtr p) 
	{
		entry = p;
	}

	// GetEntry: retrieves the current entry pointer
	inline ShadowDoorFnPtr GetPointer () 
	{
		return entry;
	}

};


// the pointer table is a container that uses a tree of nodex
// for a dictionary-based dispatch table

class PointerTable {

protected:
	Node *root;
	HANDLE hAccessLock;

	// adds an entry with a given name, by traversing the 
	// dictionary and adding nodes from the given root, 
	// and finally adding the pointer entry. 
	void AddEntry (char *name, ShadowDoorFnPtr pointer, Node *node);

	// retrieves an entry with a given name, by traversing
	// the dictionary from the specified root
	ShadowDoorFnPtr GetEntry (char *name, Node *node);

public:

	// constructor 
	PointerTable ();

	// destructor
	virtual ~PointerTable ();

	// adds an entry for a pointer with a given name
	void AddEntry (char *name, ShadowDoorFnPtr pointer);

	// retrieves a pointer with a given name
	ShadowDoorFnPtr GetEntry (char *name);
};



#endif // SD_POINTERTABLE_H