/*  *************************************************************************

	Shadow Door - a Neverwinter Nights NPC control interface
	Copyright 2003 by Robert Zubek and Phillip Saltzman, Northwestern University
    email: rob@cs.northwestern.edu, p-saltzman@northwestern.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

************************************************************************** */


#include "stdafx.h"


// extern references...
DWORD WINAPI RunListenerWrapper(LPVOID lpParam);
SOCKET initServerSock();

//
//
// singleton pointer declaration
ShadowDoor *ShadowDoor::singleton = NULL;


//
//
// protected constructor and destructor

ShadowDoor::ShadowDoor () 
{
	// _CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
	bVerbose = false;
}

ShadowDoor::~ShadowDoor () 
{
}
	

//
//
// shadow door creation:
// starts up sockets

BOOL ShadowDoor::OnCreate (const char* LogDir) 
{

	//MessageBox (NULL, "TEST", "TEST", NULL);

	// call the base class
	string log (LogDir);
	if (log.length() > 0) log += "\\";
	log += "nwnx shadow door log.txt";
	BOOL parentResult = CNWNXBase::OnCreate (log.c_str());
	if (!parentResult) return (parentResult);

	// Log banner
	Log ("Shadow Door 1.0  (%s)\n", __DATE__);
	Log ("by Phillip Saltzman and Robert Zubek\n\n");

	// set up pointer table
	handlers.AddEntry ("READY?", HandleReadyP);
	_ASSERT (handlers.GetEntry ("READY?") != NULL);
	handlers.AddEntry ("GET-TOKEN", HandleGetToken);
	handlers.AddEntry ("FINISH", HandleFinish);
	handlers.AddEntry ("SEND", HandleSend);
	handlers.AddEntry ("+V", HandleSetVerbose);
	handlers.AddEntry ("-V", HandleClearVerbose);

	//Start Winsock
	WSADATA wsData;
	WORD    wVer = MAKEWORD(2,2);
	int status = WSAStartup(wVer,&wsData);
	if (status < 0) Log ("Error initializing Winsock\n");

	//Run the lister thread
	DWORD dwThreadId;
	HANDLE hThread; 
	hThread = CreateThread(NULL, 0, RunListenerWrapper, NULL, 0, &dwThreadId);                
	CloseHandle( hThread );

	return TRUE;
}


//
//
// shadow door shutdown:
// releases sockets

BOOL ShadowDoor::OnRelease () 
{

	// clean up them sockets
	WSACleanup ();
	
	// call base class
	return CNWNXBase::OnRelease();

	// dump heap info
	_CrtDumpMemoryLeaks ();

}


//
//
// request handling:

char* ShadowDoor::OnRequest (char *gameObject, char *request, char *parameters) 
{
	if (bVerbose) Log ("SD Received request: %s\n", request);

	//ShadowDoorFnPtr foo = handlers.GetEntry ("READY?");
	ShadowDoorFnPtr fnPtr = handlers.GetEntry (request);
	if (fnPtr != NULL) 
	{
		return (this->*(fnPtr))(parameters);
	} 
	else 
	{
		Log ("* Unknown request! %s\n", request);
		return NULL;
	}
}



//////////////////////////////////////////////////////////////////////////
//
// threading and socket functions 

// Thread function to deal with incoming connections and messages.
// It loops collecting incoming characters and gathering them
// in a buffer. Once a CR is encountered, the buffer gets 
// copied and pushed onto the game-bound queue

void ShadowDoor::RunListener () 
{
	const int bufferSize = 255;
	int bufferEnd = 0;
	char buffer[bufferSize + 1];
	char elt, *s;
	int msgSize;
	Msg *pMsg;

	bConnected = false;
	servSock = initServerSock();

	Log ("Activated and waiting for connections.\n");

	// Keep opening and closing sockets
	while (1) 
	{

		//Wait for incomming connection
		externalSock = accept(servSock, NULL, NULL); 
		bConnected = true;
		Log ("Received Connection\n");
		pMsg = new Msg();

		// Read and store characters from connection. 
		do 
		{
			// read a byte at a time (inefficient? you bet! :)
			msgSize = recv (externalSock, &elt, 1, 0);
			if (msgSize == 1) {
				
				switch (elt) {

				case '#':
				case LF:
					// if it's a delimiter or end of line, cap the buffer 
					// and move it into the message currently being built
					buffer[bufferEnd] = '\0';
					s = new char [bufferEnd + 1];
					strncpy (s, buffer, bufferEnd + 1);
					if (bVerbose) Log ("Buffering message from outside world: %s\n", s);
					pMsg->push (s);
					bufferEnd = 0;

					// and if it's end of line, move the message onto queue
					// and start building a new one
					if (elt == LF) {
                        fromSocketQueue.push (pMsg);
						pMsg = new Msg();
					}
					break;

				case CR:
					// if the characters is a carriage return, ignore it!
					break;

				default:
					// otherwise it's not a special byte - 
					// add to the buffer as long as there's space
					if (bufferEnd < bufferSize)
					{
						buffer[bufferEnd++] = elt;
					}
				}
			}

		} while (msgSize > 0);

		// connection lost 
		Log ("External socket closed\n");

		bConnected = false;
		closesocket(externalSock);

		while (!pMsg->empty()) { delete pMsg->front(); pMsg->pop(); }
		delete pMsg;
   }

}


// ////////////////////////////////////////////////////////////////
//
// message handlers


// the "READY?" message is a request for data. if there's nothing
// to send to the game, it should return the string "0", otherwise
// return some other string
//
char *ShadowDoor::HandleReadyP (char *body) 
{
	if (fromSocketQueue.size() > 0) {
		return "1";
	} else {
		return "0";
	}
}

// the "GET-TOKEN" message is a request for the next available
// command token. the return value should be the token, or an
// empty string if no more are available
//
char *ShadowDoor::HandleGetToken (char *body) 
{
	char *p = "";

	if (fromSocketQueue.size() > 0 &&
		fromSocketQueue.front()->size() > 0) 
	{
		p = fromSocketQueue.front()->front();
		fromSocketQueue.front()->pop();
		stringsToDelete.push (p);
	}
	return p;
}

// the "FINISH" message informs SD that it should move on to 
// the next available command. the return value is ignored
//
char *ShadowDoor::HandleFinish (char *body) 
{
	// delete all strings on the list, and the topmost
	// message on the queue.
	while (!stringsToDelete.empty()) 
	{
		delete [] stringsToDelete.front();
		stringsToDelete.pop();
	}

	Msg *m = fromSocketQueue.front();
	fromSocketQueue.pop();
	if (!m->empty()) 
	{ 
		Log ("NB: deleting non-empty message %S ...\n", m->front());
		while (!m->empty()) { 
			delete [] m->front();
			m->pop();
		}
	}
	delete m;

	// if in debug mode, dump heap info
	_CrtDumpMemoryLeaks ();
 
	return "";
}

// the "SEND" message is a command to be sent from the game
// to the external system. the body contains the message
// and should be copied to the socket verbatim (sans terminating
// NULL), followed by a linefeed. the return value will be ignored.
//
char *ShadowDoor::HandleSend (char *body) 
{
	char lf = LF;
	int len = (int) strlen (body);
	send (externalSock, body, len, NULL);
	send (externalSock, &lf, 1, NULL);

	return "";
}


// internal handlers

char *ShadowDoor::HandleSetVerbose (char *body) 
{
	Log ("Verbose mode ON\n");
	bVerbose = true;
	return "";
}

char *ShadowDoor::HandleClearVerbose (char *body) 
{
	Log ("Verbose mode OFF\n");
	bVerbose = false;
	return "";
}